#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import re
from typing import List, Dict, Any

import libvirt

from fabric_virt_tools.utils import Utils


class NumaTune:
    def __init__(self):
        self.connection = libvirt.open(None)

    @staticmethod
    def parse_numastat(*, numastat_output: List[str]) -> Dict[Any, dict]:
        """
        Parse numa stat output and convert it to JSON
        @param numastat_output: Numa State output
        """
        '''
        Numa Stat Output looks like:
        numa_info = ["",
                 "Per-node process memory usage (in MBs) for PID 2676600 (qemu-kvm)",
                 "         Node 0 Node 1 Total",
                 "         ------ ------ -----",
                 "Private   17531  15284 32815",
                 "Heap          0      6     6",
                 "Stack         0      0     0",
                 "Huge          0      0     0",
                 "-------  ------ ------ -----",
                 "Total     17531  15290 32821"
                 ]
        '''
        result = {}
        keys = []
        for line in numastat_output:
            # Ignore empty lines and headers
            if line == "" or "Per-node" in line or "--" in line:
                continue
            # Extract Numa Nodes
            elif "Node" in line:
                line = line.strip()
                regex_pattern = r'\b(\w+\s+\d+)\b|\b(\w+)\b'
                matches = re.findall(regex_pattern, line)
                # Flatten the list of tuples and remove empty strings
                keys = [match[0] or match[1] for match in matches if match[0] or match[1]]

                # Add empty dictionary for each Node and Total
                for x in keys:
                    result[x] = {}
            else:
                # Extract Different types of Memory for each Node and Total
                regex_pattern = r'(\b\w+\b)|(\b\d+\b)'
                matches = re.findall(regex_pattern, line)
                # Flatten the list of tuples and remove empty strings
                memory_values = [match[0] or match[1] for match in matches if match[0] or match[1]]
                # Index 0 contains the type of Memory and is used as a key
                idx = 1
                for x in keys:
                    result[x][memory_values[0]] = memory_values[idx]
                    idx += 1

        return result

    @staticmethod
    def parse_numactl(*, numactl_output: List[str]) -> Dict[Any, dict]:
        """
        Parse numa stat output and convert it to JSON
        @param numastat_output: Numa State output
        """
        '''
        Numa Stat Output looks like:
        numa_info = [
                     'available: 8 nodes (0-7)', 
                     'node 0 cpus: 0 1 2 3 4 5 6 7 64 65 66 67 68 69 70 71', 
                     'node 0 size: 63798 MB', 
                     'node 0 free: 59660 MB', 
                     'node 1 cpus: 8 9 10 11 12 13 14 15 72 73 74 75 76 77 78 79', 
                     'node 1 size: 64507 MB', 
                     'node 1 free: 63042 MB', 
                     'node 2 cpus: 16 17 18 19 20 21 22 23 80 81 82 83 84 85 86 87', 
                     'node 2 size: 64507 MB', 
                     'node 2 free: 63911 MB', 
                     'node 3 cpus: 24 25 26 27 28 29 30 31 88 89 90 91 92 93 94 95', 
                     'node 3 size: 64458 MB', 
                     'node 3 free: 30882 MB', 
                     'node 4 cpus: 32 33 34 35 36 37 38 39 96 97 98 99 100 101 102 103', 
                     'node 4 size: 64507 MB', 
                     'node 4 free: 64069 MB', 
                     'node 5 cpus: 40 41 42 43 44 45 46 47 104 105 106 107 108 109 110 111', 
                     'node 5 size: 64507 MB', 
                     'node 5 free: 63966 MB', 
                     'node 6 cpus: 48 49 50 51 52 53 54 55 112 113 114 115 116 117 118 119', 
                     'node 6 size: 64507 MB', 
                     'node 6 free: 61214 MB', 
                     'node 7 cpus: 56 57 58 59 60 61 62 63 120 121 122 123 124 125 126 127', 
                     'node 7 size: 64506 MB', 
                     'node 7 free: 63869 MB', 
                     'node distances:', 
                     'node   0   1   2   3   4   5   6   7 ', 
                     '  0:  10  12  12  12  32  32  32  32 ', 
                     '  1:  12  10  12  12  32  32  32  32 ', 
                     '  2:  12  12  10  12  32  32  32  32 ', 
                     '  3:  12  12  12  10  32  32  32  32 ', 
                     '  4:  32  32  32  32  10  12  12  12 ', 
                     '  5:  32  32  32  32  12  10  12  12 ', 
                     '  6:  32  32  32  32  12  12  10  12 ', 
                     '  7:  32  32  32  32  12  12  12  10 '
                     ]
        '''
        result = {}
        for line in numactl_output:
            # Ignore empty lines and headers
            if line == "":
                continue
            elif "available" in line:
                key, value = line.split(':')
                key = key.strip()
                value = value.strip()
                result[key] = value
            elif "node distances" in line:
                break
            else:
                key, value = line.split(':')
                key = key.strip()
                value = value.strip()
                regex_pattern = r"^(.*?)\s+(\w+)$"
                matches = re.match(regex_pattern, key)
                if matches:
                    node = matches.group(1)
                    key = matches.group(2)
                    if node not in result:
                        result[node] = {}
                    result[node][key] = value
                else:
                    result[key] = value
        return result

    def __get_host_numa_info(self):
        """
        Get Host Numa Information
        :return: JSON containing NUMA information from Host
        """
        host_info_output = Utils.execute_command(command=["numactl", "-H"]).splitlines()
        return self.parse_numactl(numactl_output=host_info_output)

    def __get_guest_numa_info(self, *, domain_name: str):
        """
        Get VM Guest Numa Information
        :param domain_name: domain name for VM Guest
        :return: JSON containing NUMA information for the VM Guest
        """
        guest_info_output = Utils.execute_command(command=["numastat", "-c", "-s", domain_name]).splitlines()
        return self.parse_numastat(numastat_output=guest_info_output)

    def info(self, *, domain_name: str):
        """
        Return NUMA information for the VM guest identified by domain_name and relevant information from Host
        :param domain_name: domain name for VM Guest
        :return: SON containing NUMA information for the VM Guest and Host
        """
        host_name = self.connection.getHostname()
        result = {host_name: self.__get_host_numa_info(),
                  domain_name: self.__get_guest_numa_info(domain_name=domain_name)}
        return result

    @staticmethod
    def __extract_node_set(*, node_set: str) -> List[str]:
        """
        Extract Node Numbers from the Node Set parameter of the VM Guest
        :param node_set: Node Set parameter of the VM Guest
        :return: Node Numbers
        """
        result = []
        ranges = node_set.split(',')
        for item in ranges:
            item = item.strip()
            if item == "":
                continue
            if '-' in item:
                start, end = item.split('-')
                result.extend([str(i) for i in range(int(start), int(end) + 1)])
            else:
                result.append(item)
        return result

    def __get_allocated_memory(self, *, requested_node_set: List[str], exclude_domains: List[str] = None) -> int:
        """
        Get Already allocated memory to the VMs which have been pinned to the requested nodes
        :param requested_node_set: Requested Node Set
        :param exclude_domains: List of domain names to exclude
        :return:
        """
        result = 0
        # Traverse through all the running VMs
        for domain_id in self.connection.listDomainsID():
            domain = self.connection.lookupByID(domain_id)

            # Exclude any domains in the exclude list
            if exclude_domains is not None and domain.name() in exclude_domains:
                continue

            # Determine the Numa Node Set assigned to the domain
            domain_numa_nodeset = self.__extract_node_set(node_set=domain.numaParameters().get('numa_nodeset'))

            # Find common elements between assigned and requested node set
            common_elements = list(set(domain_numa_nodeset).intersection(requested_node_set))
            if len(common_elements) > 0:
                # Get Guest Numa Infor
                domain_numa_info = self.__get_guest_numa_info(domain_name=domain.name())

                # Determine the allocated memory on the requested Numa Node
                for n in common_elements:
                    node_numa_allocated = domain_numa_info.get(f"Node {n}").get("Total")
                    result += int(node_numa_allocated)

        return result

    def numa_tune(self, *, domain_name: str, node_set: List[str]):
        """
        Tune Numa for VM guest to the requested nodes
        :param domain_name: Domain name for the VM
        :param node_set: Requested Node Set
        :return:
        """
        try:
            domain = self.connection.lookupByName(domain_name)
            # Unit is Bytes
            requested_memory = domain.maxMemory()
            # Convert to MB
            requested_memory /= 1024

            # Get Allocated memory already in use for the requested node set
            allocated_memory = self.__get_allocated_memory(requested_node_set=node_set, exclude_domains=[domain_name])

            # Get Host Numa Info
            host_numa = self.__get_host_numa_info()

            # Determine Total Memory for the Requested Numa Nodes
            total_memory = 0
            for node in node_set:
                node_info = host_numa.get(f"node {node}")
                total_memory_str = node_info.get("size").replace(" MB", "")
                total_memory += int(total_memory_str)

            # Verify requested memory can still be accommodated by requested Numa Nodes
            if requested_memory <= (total_memory - allocated_memory):
                Utils.execute_command(["virsh", "numatune", domain_name, "--nodeset", ','.join(node_set),
                                      "--live", "--config"])
            else:
                raise Exception(f"NUMA_TUNE_ERROR - Not enough memory available for {domain_name} on Numa Node: {node_set}")
        except libvirt.libvirtError as e:
            raise Exception(f"NUMA_TUNE_ERROR for {domain_name} e: {e}")


if __name__ == '__main__':
    numa_tune = NumaTune()
    print(numa_tune.info(domain_name='instance-0000106b'))
    numa_tune.numa_tune(domain_name='instance-0000106b', node_set=["5", "6"])




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import json
import subprocess
from typing import List, Dict

import libvirt

from fabric_virt_tools.utils import Utils


class CpuTune:
    """
    Supports CPU pinning for VM Guests and retrieving vCpuInfo
    """
    VCPU = "vcpu"
    CPU = "cpu"
    NUMA = "numa"
    FIELD = "field"
    DATA = "data"
    PINNED_CPUS = "pinned_cpus"
    CPU_AFFINITY = "CPU Affinity"

    def __init__(self):
        """
        Constructor
        """
        # Open a connection with libvirt
        self.connection = libvirt.open(None)

    def __get_pinned_cpus(self, *, exclude_domains: List[str] = None) -> List[str]:
        """
        Check all the running VMs and determine the host CPUs which are pinned
        :param exclude_domains: List of domain names to exclude
        :return: list of host CPUs which are pinned
        """
        pinned_cpu_list = []
        # Traverse through currently active VMs
        for domain_id in self.connection.listDomainsID():
            domain = self.connection.lookupByID(domain_id)

            # Exclude the domain if included in exclude list
            if exclude_domains is not None and domain.name() in exclude_domains:
                continue
            pins = domain.vcpuPinInfo()
            for (i, cpu_map) in enumerate(pins):
                # cpu_map is an bitfield (list of bools) with the length of available CPU cores
                # If a CPU is pinned, exactly one value is True, its index is indicating which
                # CPU we are talking about
                if cpu_map.count(True) == 1:
                    cpus = [str(i) for i, x in enumerate(cpu_map) if x]
                    for c in cpus:
                        pinned_cpu_list.append(c)
        return pinned_cpu_list

    def vcpu_pin(self, *, domain_name: str, vcpu_cpu_map: List[Dict[str, str]]):
        """
        Pin vCpus of  VM to specific Host Cpus
        :param domain_name: Domain Name for the VM Guest
        :param vcpu_cpu_map: Virtual CPU to Host CPU Map
        :return:
        """
        # Determine already pinned host cpus
        current_pinned_cpus = self.__get_pinned_cpus(exclude_domains=[domain_name])
        # Verify that the requested CPUs are not already pinned
        for vcpu_cpu in vcpu_cpu_map:
            if vcpu_cpu.get(CpuTune.CPU) in current_pinned_cpus:
                raise Exception(f"CPU_PIN_ERROR: Requested CPU: {vcpu_cpu} is already pinned to another domain")
        try:
            # Grab the Domain for the VM
            domain = self.connection.lookupByName(domain_name)
            # Determine the current CPU Map
            pins = domain.vcpuPinInfo()

            # Build new CPU Map based on the VCPU to Host CPU map requested for each VCPU
            for vcpu_cpu in vcpu_cpu_map:
                vcpu = int(vcpu_cpu.get(CpuTune.VCPU))
                cpu = int(vcpu_cpu.get(CpuTune.CPU))
                cpu_map = pins[vcpu]
                cpu_map = tuple([True if j == cpu else False for j in range(len(cpu_map))])
                # Pin vCPU to the requested Host CPU
                domain.pinVcpuFlags(vcpu, cpu_map, libvirt.VIR_DOMAIN_AFFECT_CONFIG|libvirt.VIR_DOMAIN_AFFECT_LIVE)
        except libvirt.libvirtError as e:
            raise Exception(f"CPU_PIN_ERROR for {domain_name} e: {e}")

    @staticmethod
    def __get_host_cpu_info():
        """
        Get Host CPU Information using "lscpu -J" command
        :return: JSON object containing CPU and Numa parameters returned by lscpu
        """
        try:
            # Execute the command
            output = Utils.execute_command(["lscpu", "-J"])
            output_json = json.loads(output)["lscpu"]
            # Extract the output
            host = {}
            # Traverse the list
            for x in output_json:
                key = x[CpuTune.FIELD]
                value = x[CpuTune.DATA].replace(":", "")
                # Extract CPU/Numa parameters
                if CpuTune.CPU.upper() in key or CpuTune.NUMA in key:
                    host[key] = value
            return host
        except subprocess.CalledProcessError as e:
            return None

    @staticmethod
    def parse_vcpuinfo(*, vcpu_info_output: List[str]) -> List[Dict[str, str]]:
        """
        Parse vCpu Info and transform it to JSON
        @param vcpu_info_output: vCpu Info command output
        """
        '''
        Vcpuinfo Output looks like:
            cpu_info = ["VCPU:           0",
                "CPU:            27",
                "State:          running",
                "CPU time:       3.1s",
                "CPU Affinity:   yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy", 
                "", 
                "VCPU:           1", 
                "CPU:            9", 
                "State:          running", 
                "CPU Affinity:   yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy"
                ]
        '''
        result = []
        data = {}
        for line in vcpu_info_output:
            line = line.strip()
            if line == "":
                result.append(data.copy())
                data.clear()
            else:
                # Split the line at the first occurrence of ':' character
                key, value = line.split(':', 1)
                # Remove leading/trailing spaces from key and value
                key = key.strip()
                value = value.strip()
                # Add the key-value pair to the data dictionary
                data[key] = value
        return result

    def info(self, *, domain_name: str):
        """
        Determine CPU Info for the VM Guest and the host
        :param domain_name:
        :return:
        """
        # Determine Host Name
        host_name = self.connection.getHostname()
        # Get Host CPU Information
        host_info = self.__get_host_cpu_info()
        # Get Pinned CPUs
        pinned_cpus = self.__get_pinned_cpus(exclude_domains=[domain_name])
        host_info[CpuTune.PINNED_CPUS] = pinned_cpus
        # Get vCPU Information
        vcpu_info_output = Utils.execute_command(command=["virsh", "vcpuinfo", domain_name]).splitlines()
        vcpu_info = self.parse_vcpuinfo(vcpu_info_output=vcpu_info_output)

        try:
            domain = self.connection.lookupByName(domain_name)
            pins = domain.vcpuPinInfo()
            # Update CPU Affinity for each vCPU
            for (i, cpu_map) in enumerate(pins):
                cpus = [str(i) for i, x in enumerate(cpu_map) if x]
                vcpu_info[i][CpuTune.CPU_AFFINITY] = ','.join(cpus)
        except libvirt.libvirtError as e:
            raise Exception(f"CPU_PIN_ERROR for {domain_name} e: {e}")

        # Return Host and Guest vCPU Info
        result = {host_name: host_info,
                  domain_name: vcpu_info}
        return result


if __name__ == '__main__':
    cpu_tune = CpuTune()
    cpu_tune.vcpu_pin(domain_name='instance-0000106b', vcpu_cpu_map=[{"vcpu": '0',"cpu": '112'},
                                                                     {"vcpu": '1',"cpu": '113'}])

    print(cpu_tune.info(domain_name='instance-0000106b'))
